@@ -3,87 +3,31 @@
 from __future__ import annotations
 
 import warnings
-from collections.abc import Callable, Hashable, MutableMapping
+from collections.abc import Hashable, MutableMapping
 from functools import partial
 from typing import TYPE_CHECKING, Any, Union
 
 import numpy as np
 import pandas as pd
 
+from xarray.coding.common import (
+    SerializationWarning,
+    VariableCoder,
+    lazy_elemwise_func,
+    pop_to,
+    safe_setitem,
+    unpack_for_decoding,
+    unpack_for_encoding,
+)
+from xarray.coding.times import CFDatetimeCoder, CFTimedeltaCoder
 from xarray.core import dtypes, duck_array_ops, indexing
 from xarray.core.variable import Variable
-from xarray.namedarray.parallelcompat import get_chunked_array_type
-from xarray.namedarray.pycompat import is_chunked_array
 
 if TYPE_CHECKING:
     T_VarTuple = tuple[tuple[Hashable, ...], Any, dict, dict]
     T_Name = Union[Hashable, None]
 
 
-class SerializationWarning(RuntimeWarning):
-    """Warnings about encoding/decoding issues in serialization."""
-
-
-class VariableCoder:
-    """Base class for encoding and decoding transformations on variables.
-
-    We use coders for transforming variables between xarray's data model and
-    a format suitable for serialization. For example, coders apply CF
-    conventions for how data should be represented in netCDF files.
-
-    Subclasses should implement encode() and decode(), which should satisfy
-    the identity ``coder.decode(coder.encode(variable)) == variable``. If any
-    options are necessary, they should be implemented as arguments to the
-    __init__ method.
-
-    The optional name argument to encode() and decode() exists solely for the
-    sake of better error messages, and should correspond to the name of
-    variables in the underlying store.
-    """
-
-    def encode(self, variable: Variable, name: T_Name = None) -> Variable:
-        """Convert an encoded variable to a decoded variable"""
-        raise NotImplementedError()
-
-    def decode(self, variable: Variable, name: T_Name = None) -> Variable:
-        """Convert an decoded variable to a encoded variable"""
-        raise NotImplementedError()
-
-
-class _ElementwiseFunctionArray(indexing.ExplicitlyIndexedNDArrayMixin):
-    """Lazily computed array holding values of elemwise-function.
-
-    Do not construct this object directly: call lazy_elemwise_func instead.
-
-    Values are computed upon indexing or coercion to a NumPy array.
-    """
-
-    def __init__(self, array, func: Callable, dtype: np.typing.DTypeLike):
-        assert not is_chunked_array(array)
-        self.array = indexing.as_indexable(array)
-        self.func = func
-        self._dtype = dtype
-
-    @property
-    def dtype(self) -> np.dtype:
-        return np.dtype(self._dtype)
-
-    def _oindex_get(self, key):
-        return type(self)(self.array.oindex[key], self.func, self.dtype)
-
-    def _vindex_get(self, key):
-        return type(self)(self.array.vindex[key], self.func, self.dtype)
-
-    def __getitem__(self, key):
-        return type(self)(self.array[key], self.func, self.dtype)
-
-    def get_duck_array(self):
-        return self.func(self.array.get_duck_array())
-
-    def __repr__(self) -> str:
-        return f"{type(self).__name__}({self.array!r}, func={self.func!r}, dtype={self.dtype!r})"
-
-
 class NativeEndiannessArray(indexing.ExplicitlyIndexedNDArrayMixin):
     """Decode arrays on the fly from non-native to native endianness
 
@@ -161,63 +105,6 @@ class BoolTypeArray(indexing.ExplicitlyIndexedNDArrayMixin):
         return np.asarray(self.array[key], dtype=self.dtype)
 
 
-def lazy_elemwise_func(array, func: Callable, dtype: np.typing.DTypeLike):
-    """Lazily apply an element-wise function to an array.
-    Parameters
-    ----------
-    array : any valid value of Variable._data
-    func : callable
-        Function to apply to indexed slices of an array. For use with dask,
-        this should be a pickle-able object.
-    dtype : coercible to np.dtype
-        Dtype for the result of this function.
-
-    Returns
-    -------
-    Either a dask.array.Array or _ElementwiseFunctionArray.
-    """
-    if is_chunked_array(array):
-        chunkmanager = get_chunked_array_type(array)
-
-        return chunkmanager.map_blocks(func, array, dtype=dtype)  # type: ignore[arg-type]
-    else:
-        return _ElementwiseFunctionArray(array, func, dtype)
-
-
-def unpack_for_encoding(var: Variable) -> T_VarTuple:
-    return var.dims, var.data, var.attrs.copy(), var.encoding.copy()
-
-
-def unpack_for_decoding(var: Variable) -> T_VarTuple:
-    return var.dims, var._data, var.attrs.copy(), var.encoding.copy()
-
-
-def safe_setitem(dest, key: Hashable, value, name: T_Name = None):
-    if key in dest:
-        var_str = f" on variable {name!r}" if name else ""
-        raise ValueError(
-            f"failed to prevent overwriting existing key {key} in attrs{var_str}. "
-            "This is probably an encoding field used by xarray to describe "
-            "how a variable is serialized. To proceed, remove this key from "
-            "the variable's attributes manually."
-        )
-    dest[key] = value
-
-
-def pop_to(
-    source: MutableMapping, dest: MutableMapping, key: Hashable, name: T_Name = None
-) -> Any:
-    """
-    A convenience function which pops a key k from source to dest.
-    None values are not passed on.  If k already exists in dest an
-    error is raised.
-    """
-    value = source.pop(key, None)
-    if value is not None:
-        safe_setitem(dest, key, value, name=name)
-    return value
-
-
 def _apply_mask(
     data: np.ndarray,
     encoded_fill_values: list,
@@ -234,6 +121,8 @@ def _apply_mask(
 
 def _is_time_like(units):
     # test for time-like
+    # return "datetime" for datetime-like
+    # return "timedelta" for timedelta-like
     if units is None:
         return False
     time_strings = [
@@ -255,9 +144,9 @@ def _is_time_like(units):
             _unpack_netcdf_time_units(units)
         except ValueError:
             return False
-        return True
+        return "datetime"
     else:
-        return any(tstr == units for tstr in time_strings)
+        return "timedelta" if any(tstr == units for tstr in time_strings) else False
 
 
 def _check_fill_values(attrs, name, dtype):
@@ -367,6 +256,14 @@ def _encode_unsigned_fill_value(
 class CFMaskCoder(VariableCoder):
     """Mask or unmask fill values according to CF conventions."""
 
+    def __init__(
+        self,
+        decode_times: bool | CFDatetimeCoder = False,
+        decode_timedelta: bool | CFTimedeltaCoder = False,
+    ) -> None:
+        self.decode_times = decode_times
+        self.decode_timedelta = decode_timedelta
+
     def encode(self, variable: Variable, name: T_Name = None):
         dims, data, attrs, encoding = unpack_for_encoding(variable)
 
@@ -393,22 +290,33 @@ class CFMaskCoder(VariableCoder):
 
         if fv_exists:
             # Ensure _FillValue is cast to same dtype as data's
+            # but not for packed data
             encoding["_FillValue"] = (
                 _encode_unsigned_fill_value(name, fv, dtype)
                 if has_unsigned
-                else dtype.type(fv)
+                else (
+                    dtype.type(fv)
+                    if "add_offset" not in encoding and "scale_factor" not in encoding
+                    else fv
+                )
             )
             fill_value = pop_to(encoding, attrs, "_FillValue", name=name)
 
         if mv_exists:
             # try to use _FillValue, if it exists to align both values
             # or use missing_value and ensure it's cast to same dtype as data's
+            # but not for packed data
             encoding["missing_value"] = attrs.get(
                 "_FillValue",
                 (
                     _encode_unsigned_fill_value(name, mv, dtype)
                     if has_unsigned
-                    else dtype.type(mv)
+                    else (
+                        dtype.type(mv)
+                        if "add_offset" not in encoding
+                        and "scale_factor" not in encoding
+                        else mv
+                    )
                 ),
             )
             fill_value = pop_to(encoding, attrs, "missing_value", name=name)
@@ -416,19 +324,37 @@ class CFMaskCoder(VariableCoder):
         # apply fillna
         if fill_value is not None and not pd.isnull(fill_value):
             # special case DateTime to properly handle NaT
-            if _is_time_like(attrs.get("units")) and data.dtype.kind in "iu":
-                data = duck_array_ops.where(
-                    data != np.iinfo(np.int64).min, data, fill_value
-                )
+            if _is_time_like(attrs.get("units")):
+                if data.dtype.kind in "iu":
+                    data = duck_array_ops.where(
+                        data != np.iinfo(np.int64).min, data, fill_value
+                    )
+                else:
+                    # if we have float data (data was packed prior masking)
+                    # we just fillna
+                    data = duck_array_ops.fillna(data, fill_value)
+                    # but if the fill_value is of integer type
+                    # we need to round and cast
+                    if np.array(fill_value).dtype.kind in "iu":
+                        data = duck_array_ops.astype(
+                            duck_array_ops.around(data), type(fill_value)
+                        )
             else:
                 data = duck_array_ops.fillna(data, fill_value)
 
         if fill_value is not None and has_unsigned:
             pop_to(encoding, attrs, "_Unsigned")
             # XXX: Is this actually needed? Doesn't the backend handle this?
-            signed_dtype = np.dtype(f"i{dtype.itemsize}")
-            data = duck_array_ops.view(
-                duck_array_ops.astype(duck_array_ops.around(data), signed_dtype), dtype
+            # two-stage casting to prevent undefined cast from float to unsigned int
+            # first float -> int with corresponding itemsize
+            # second int -> int/uint to final itemsize
+            signed_dtype = np.dtype(f"i{data.itemsize}")
+            data = duck_array_ops.astype(
+                duck_array_ops.astype(
+                    duck_array_ops.around(data), signed_dtype, copy=False
+                ),
+                dtype,
+                copy=False,
             )
             attrs["_FillValue"] = fill_value
 
@@ -460,19 +386,28 @@ class CFMaskCoder(VariableCoder):
             )
 
         if encoded_fill_values:
-            # special case DateTime to properly handle NaT
             dtype: np.typing.DTypeLike
             decoded_fill_value: Any
-            if _is_time_like(attrs.get("units")) and data.dtype.kind in "iu":
-                dtype, decoded_fill_value = np.int64, np.iinfo(np.int64).min
+            # in case of packed data we have to decode into float
+            # in any case
+            if "scale_factor" in attrs or "add_offset" in attrs:
+                dtype, decoded_fill_value = (
+                    _choose_float_dtype(data.dtype, attrs),
+                    np.nan,
+                )
             else:
-                if "scale_factor" not in attrs and "add_offset" not in attrs:
-                    dtype, decoded_fill_value = dtypes.maybe_promote(data.dtype)
+                # in case of no-packing special case DateTime/Timedelta to properly
+                # handle NaT, we need to check if time-like will be decoded
+                # or not in further processing
+                is_time_like = _is_time_like(attrs.get("units"))
+                if (
+                    (is_time_like == "datetime" and self.decode_times)
+                    or (is_time_like == "timedelta" and self.decode_timedelta)
+                ) and data.dtype.kind in "iu":
+                    dtype = np.int64
+                    decoded_fill_value = np.iinfo(np.int64).min
                 else:
-                    dtype, decoded_fill_value = (
-                        _choose_float_dtype(data.dtype, attrs),
-                        np.nan,
-                    )
+                    dtype, decoded_fill_value = dtypes.maybe_promote(data.dtype)
 
             transform = partial(
                 _apply_mask,
@@ -552,6 +487,14 @@ class CFScaleOffsetCoder(VariableCoder):
         decode_values = encoded_values * scale_factor + add_offset
     """
 
+    def __init__(
+        self,
+        decode_times: bool | CFDatetimeCoder = False,
+        decode_timedelta: bool | CFTimedeltaCoder = False,
+    ) -> None:
+        self.decode_times = decode_times
+        self.decode_timedelta = decode_timedelta
+
     def encode(self, variable: Variable, name: T_Name = None) -> Variable:
         dims, data, attrs, encoding = unpack_for_encoding(variable)
 
@@ -581,11 +524,16 @@ class CFScaleOffsetCoder(VariableCoder):
                 scale_factor = np.asarray(scale_factor).item()
             if np.ndim(add_offset) > 0:
                 add_offset = np.asarray(add_offset).item()
-            # if we have a _FillValue/masked_value we already have the wanted
+            # if we have a _FillValue/masked_value in encoding we already have the wanted
             # floating point dtype here (via CFMaskCoder), so no check is necessary
-            # only check in other cases
+            # only check in other cases and for time-like
             dtype = data.dtype
-            if "_FillValue" not in encoding and "missing_value" not in encoding:
+            is_time_like = _is_time_like(attrs.get("units"))
+            if (
+                ("_FillValue" not in encoding and "missing_value" not in encoding)
+                or (is_time_like == "datetime" and self.decode_times)
+                or (is_time_like == "timedelta" and self.decode_timedelta)
+            ):
                 dtype = _choose_float_dtype(dtype, encoding)
 
             transform = partial(
